from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from httplib import HTTPConnection
from lxml.etree import Element, ElementTree, SubElement, fromstring, tostring, parse
from merescocore.components import readConfig
from sys import argv
from time import ctime
from urllib import urlopen
from xml.sax.saxutils import unescape

####
## This gateway handles sruUpdates of a Meresco harvester (or any other server that sends sruUpdates).
## There are two kinds of updates: adding jel-codes to eo-records and adding ddi-documents to eo-records
## The sruUpdates have to be processed by the gateway to make them suitable as sruUpdates that can be sent to a
## a (Meresco) writer
####

diag_ns = 'http://www.loc.gov/zing/srw/diagnostic/'
meta_ns = 'http://meresco.com/namespace/harvester/meta'
mods_ns = 'http://www.loc.gov/mods/v3'
srw_ns  = 'http://www.loc.gov/zing/srw/'
ucp_ns  = 'info:lc/xmlns/update-v1'
ns = {'ddi':'http://www.icpsr.umich.edu/DDI',
      'diag': diag_ns,
      'didl': 'urn:mpeg:mpeg21:2002:02-DIDL-NS',
      'dii' : 'urn:mpeg:mpeg21:2002:01-DII-NS',
      'meta': meta_ns,
      'mods': mods_ns,
      'srw' : srw_ns,
      'ucp' : ucp_ns}


class SruUpdateHandler(BaseHTTPRequestHandler):
    def doRequest(self, query, xml, part=""):
        """Preparing and sending a request to a remote writer.

        A sruUpdate request from a harvester can result in more than one sruUpdate requests to a writer.
        For each request this function has to be called.
        query: used in sru request that searches for the publication that must be enriched
        xml: object with the enrichment info as supplied by the sender of the sruUpdate request (harvester).
        part: name of part that must be added to the record
        """
        try:
            sruResponse = \
                parse(urlopen(config['baseurl'] + \
                "?version=1.1&operation=searchRetrieve&recordSchema=meta&x-recordSchema=metadata&query=" + query))
        except IOError, (errno, strerror):
            self._status = "fail"
            self._details += " baseurl=%s errno=%s strerror=%s" % (baseurl, errno, strerror)
        else:
            #print tostring(sruResponse, pretty_print=True)
            if sruResponse.xpath("//srw:numberOfRecords/text()", ns)[0] == '0':
                ## no publications found ##
                self._status = "fail"                
                self._details += " reason=record not found in remote database"
            else:
                ## publication found. FIXME what happens when # hits > 1?
                ## get info from meta-part ##
                collection = sruResponse.xpath("//meta:meta/meta:repository/meta:collection/text()", ns)[0]
                repositoryGroupId = sruResponse.xpath("//meta:meta/meta:repository/meta:repositoryGroupId/text()", ns)[0]
                repository = sruResponse.xpath("//meta:meta/meta:repository/meta:id/text()", ns)[0]
                recordIdentifier = sruResponse.xpath("//meta:meta/meta:record/meta:id/text()", ns)[0]
                parts = sruResponse.xpath("//meta:meta/meta:parts/meta:part/text()", ns)
                self._details += " uploadId=" + repository + ":" + recordIdentifier

                ## new Meta object ##
                meta = Meta(collection, repositoryGroupId, repository, recordIdentifier, parts, part).get()

                ## creating new Document object ##
                if part == "ddi":
                    document = Document(meta, tostring(xml), part=part).get(method="Element")
                elif part == "jel":
                    jels = xml.xpath\
                        ("//didl:DIDL/didl:Item/didl:Component/didl:Resource/mods:mods/mods:classification[@authority='info:eu-repo/classification/jel']/text()", ns)
                    if jels:
                        mods = Mods(jels).get()
                        #print mods
                        document = Document(meta, mods, part="jel").get(method="Element")
                    else:
                        self._status = "fail"
                        self._details += " reason=no jel-codes supplied"

                ## when there is a new Document, it can be sent to the appropriate writer ##
                if document:
                    recordUpdate = UpdateRequest("replace", repository, recordIdentifier, document)
                    #print recordUpdate.get(xml_declaration=True)

                    ## determine port of writer
                    port = config['port_default_writer']
                    if collection == "repec": port = config['port_repec']
                    elif collection == "eo-repo": port = config['port_eo-repo']
                    self._details += " wport=" + port
                    
                    self._status, d = recordUpdate.send(port)
                    if d: self._details += " diagnostic=" + d

        finally:
            updateResponse = UpdateResponse(self._status, details=self._details, message=self._message).get(xml_declaration=True)
            #print updateResponse
            print self._status, self._details
            logfile.write(ctime() + " " + self._status + self._details + "\n")
            self.wfile.write(updateResponse)

    def do_POST(self):
        self._status = "fail"
        self._details = ""
        self._message = ""
        try:
            updateRequest = self.rfile.read(int(self.headers['Content-Length']))
            self.send_response(200)
            self.send_header("Content-Type", "text/xml, charset=utf-8")
            self.end_headers()
  
            xml = fromstring(unescape(updateRequest))
            #print tostring(xml)

            codeBook = xml.xpath("//ddi:codeBook", ns)
            if codeBook:
                codeBook = codeBook[0]
                #print tostring(codeBook, pretty_print=True)
                relPubls = codeBook.xpath("//ddi:codeBook/ddi:stdyDscr/ddi:othrStdyMat/ddi:relPubl/text()", ns)
                for relPubl in relPubls:
                    ## uit relPub identifier van publicatie halen ##
                    #print relPubl
                    # toch maar eens een reguliere expressie gebruiken?
                    eo = "economistsonline.org/publications/?id="
                    loceo = relPubl.find(eo)
                    if loceo < 0 :
                        eo = "economistsonline.org/publications?id="
                        loceo = relPubl.find(eo)
                    if loceo < 0: continue
                    s = relPubl[loceo + len(eo):]
                    # single quotes?
                    locquote = s.find('\"')
                    identifier = ""
                    if locquote > 0: identifier = s[:locquote]
                    else: identifier = s
                    identifier = identifier.strip()
                    #print "identifier = " + identifier + "|"
                    self._details += " type=ddi pubid=" + identifier 
                    self.doRequest(identifier, codeBook, part="ddi")
           
            DIDL = xml.xpath("//didl:DIDL", ns)
            if DIDL:
                DIDL = DIDL[0]
                #print tostring(DIDL, pretty_print=True)
                docid = DIDL.xpath("//didl:DIDL/didl:Descriptor/didl:Statement/dii:Identifier/text()", ns)[0]
                self._details += " type=jel docid=" + docid
                #print docid
                query = "norm.docid%20exact%20%22" + docid + "%22"
                self.doRequest(query, DIDL, part="jel" )
        except:
            pass

class Tree(object):
    def __init__(self):
       self.root = None

    def get(self, method="xml", xml_declaration=False):
        if method == "Element":
            return self.root
        elif method == "xml":
            return tostring(self.root, xml_declaration=xml_declaration, encoding="utf-8")

class Mods(Tree):
    def __init__(self, jels):
        self.ns = {'xmlns:mods': mods_ns}
        self.root = Element('mods:mods', self.ns, version="3.2")
        for jel in jels:
             SubElement(self.root, 'mods:classification', authority="info:eu-repo/classification/jel").text = jel

class Meta(Tree):
    def __init__(self, collection, repositoryGroupId, repository, recordIdentifier, parts, partName):
        self.ns = {'xmlns:meta': meta_ns}
        self.root = Element('meta:meta', self.ns)
        self.upload = SubElement(self.root, 'meta:upload')
        SubElement(self.upload, 'meta:id').text = repository + ":" + recordIdentifier
        self.record =  SubElement(self.root, 'meta:record')
        SubElement(self.record, 'meta:id').text = recordIdentifier
        self.repository = SubElement(self.root, 'meta:repository')
        SubElement(self.repository, 'meta:id').text = repository
        SubElement(self.repository, 'meta:repositoryGroupId').text = repositoryGroupId
        SubElement(self.repository, 'meta:collection').text = collection
        self.parts = SubElement(self.root, 'meta:parts')
        SubElement(self.parts, 'meta:part').text = partName
        if parts:
            for part in parts:
                if part == partName: continue
                SubElement(self.parts, 'meta:part').text = part

class Document(Tree):
    def __init__(self, meta, record, part=""):
        self.root = Element('document', xmlns="http://meresco.com/namespace/harvester/document")
        SubElement(self.root, 'part', name="meta").text = meta
        SubElement(self.root, 'part', name=part).text = record 

class UpdateResponse(Tree):
    def __init__(self, status, details="", message=""):
        self.ns = {"xmlns:srw": srw_ns, "xmlns:ucp": ucp_ns}
        self.root = Element('srw:updateResponse', self.ns)
        SubElement(self.root, 'srw:version').text = "1.0"
        SubElement(self.root, 'ucp:operationStatus').text = status
        if status == "fail":
            self.ns = {"xmlns:diag": diag_ns}
            self.diag = SubElement(self.root, 'diag:diagnostic', self.ns)
            SubElement(self.diag, 'diag:uri').text = "info:srw/diagnostic/12/1"
            SubElement(self.diag, 'diag:details').text = details
            SubElement(self.diag, 'diag:message').text = "Invalid component: record rejected"

class UpdateRequest(Tree):
    def __init__(self, action, repository, recordIdentifier, document):
        self.ns = {"xmlns:srw": srw_ns, "xmlns:ucp": ucp_ns}
        self.root = Element('srw:updateRequest', self.ns)
        SubElement(self.root, 'srw:version').text = "1.0"
        self.action = SubElement(self.root, 'ucp:action')
        if action == "replace" : self.action.text = "info:srw/action/1/replace"
        elif action == "delete": self.action.text = "info:srw/action/1/delete"
        SubElement(self.root, 'ucp:recordIdentifier').text = repository + ":" + recordIdentifier
        if action == "replace":
            self.record = SubElement(self.root, 'srw:record')
            SubElement(self.record, 'srw:recordPacking').text = 'xml'
            SubElement(self.record, 'srw:recordSchema').text  = 'document'
            SubElement(self.record, 'srw:recordData').append(document)

    def send(self, port):
        # FIXME HTTPExceptions afvangen
        self.request = self.get(xml_declaration=True)
        #print self.request
        ## writer waarnaar de enrichment wordt gestuurd ##
        connection = HTTPConnection('localhost', port)
        connection.putrequest("POST", "/update")
        connection.putheader("Content-Type", "text/xml; charset=\"utf-8\"")
        connection.putheader("Content-Length", str(len(self.request)))
        connection.endheaders()
        connection.send(self.request)

        response = connection.getresponse()
        status = response.status
        content = fromstring(response.read())
        #print tostring(content)
        if status == 200:
            if content.xpath("//ucp:operationStatus/text()", ns)[0] == "success":
                return "success", ""
            else:
                return "fail", content.xpath("//diag:diagnostic/diag:details/text()", ns)[0]
        else:
            return "fail", "HTTP problem: %s" %status

def main():
    args=argv[1:]
    if len(args) != 1: 
        print "supply one argument for the config file"
        exit()
    else:
        # FIXME mooi is anders, maar ik weet niet beter.
        # BaseHTTPRequestHandler heeft een instance variable server die gebruikt kan worden,
        # maar ik weet niet hoe.
        global config
        config = readConfig(args[0])
        # FIXME alternatief is om de log_* methods van BaseHTTPRequestHandler te gebruiken
        global logfile
        logfile = open(config['logfile'], "a", 1)
    try:
        port = int(config['port_server'])
        server = HTTPServer(('', port), SruUpdateHandler)
        print 'started httpserver...'
        logfile.write(ctime() + " started httpserver\n")
        server.serve_forever()
    except KeyboardInterrupt:
        print '^C received, shutting down server'
        logfile.close()
        server.socket.close()

if __name__ == '__main__':
    main()
